"""
Interfaces used by the Workspaces' capabilities system.
"""

import inspect
from abc import ABC, abstractmethod
from typing import List, Union, Optional

from .product_interfaces import ProductIF, FutureProductIF

ProductLocator = str
CapabilityName = str


class CapabilityIF:
    @property
    def id(self) -> int: raise NotImplementedError

    @property
    def name(self) -> str: raise NotImplementedError

    @property
    def steps(self) -> List["CapabilityStepIF"]: raise NotImplementedError

    @property
    def max_jobs(self) -> int: raise NotImplementedError

    def create_request(self, parameters: "ParameterIF", future_products: FutureProductIF) -> "CapabilityRequestIF":
        raise NotImplementedError


class CapabilityRequestIF:
    @property
    def id(self) -> int: raise NotImplementedError

    @property
    def state(self) -> str: raise NotImplementedError

    @property
    def capability(self) -> CapabilityIF: raise NotImplementedError

    @property
    def parameters(self) -> str: raise NotImplementedError

    @property
    def future_products(self) -> str: raise NotImplementedError

    @property
    def versions(self) -> List["CapabilityVersionIF"]: raise NotImplementedError


class CapabilityExecutionIF:
    pass


"""

Capability System / Services

"""


class CapabilityInfoIF(ABC):
    """
    Data access object that can look up and record information about capabilities
    and capability requests.
    """
    @abstractmethod
    def create_capability(
            self,
            name: CapabilityName,
            steps: List["CapabilityStepType"],
            max_jobs: int
    ) -> int:
        """
        Create new capability and save it in the database
        :param name: Name of new capability
        :param steps: List of capability steps required to execute the capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :return: Integer identifier for the capability
        """
        raise NotImplementedError

    @abstractmethod
    def create_capability_request(
            self,
            capability_name: str,
            parameters: List["ParameterIF"] = None,
            future_products: List[FutureProductIF] = None,
            versions: List[str] = None
    ) -> int:
        """
        Create new capability request and save it in the database
        :param capability_id: ID of the requested capability
        :param parameters: List
        :param future_products:
        :param versions:
        :return: Integer identifier of the request
        """
        raise NotImplementedError

    @abstractmethod
    def create_execution_record(self, request_id: int) -> int:
        """
        Create new execution record for a request and save it in the database
        :param request_id: ID of the capability request
        :return: Integer identifier for the record
        """
        raise NotImplementedError

    @abstractmethod
    def lookup_capability(self, capability_name: str) -> CapabilityIF:
        raise NotImplementedError

    @abstractmethod
    def lookup_entity(
            self,
            entity_id: int,
            entity_schema: Union[CapabilityIF, CapabilityRequestIF, CapabilityExecutionIF]
    ) -> Optional[Union[CapabilityIF, CapabilityRequestIF, CapabilityExecutionIF]]:
        """
        Look up entity in database and return object representation of it if found
        :param entity_id: ID of entity to be searched for
        :param entity_schema: Database schema of the entity
        :return: Object representation of entity if found, else None
        """
        raise NotImplementedError

    @abstractmethod
    def save_entity(self, entity: Union[CapabilityIF, CapabilityRequestIF, CapabilityExecutionIF]) -> int:
        """
        Save a given entity and return an integer identifier for it
        :param entity: The entity to save
        :return: The entity's identifier
        """
        raise NotImplementedError()


class ProjectSettingsIF(ABC):
    """
    Class that deals with projects overriding capability settings and making multiple instances
    of a capability with the same name
    """
    pass


class CapabilityMatrixIF(ABC):
    """
    Maps CASA versions to version-specific templates, allowing us to support a variety of
    CASA versions
    """
    pass


class CapabilityQueueIF(ABC):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """
    @abstractmethod
    def enqueue(self, request: CapabilityRequestIF, priority: int):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def dequeue(self):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def pause(self):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


class CapabilityEngineIF(ABC):
    """
    Executes a prepare and run workflow step of a capability
    """
    @abstractmethod
    def execute(self):
        pass


class CapabilityServiceIF(ABC):
    """
    The capability service: clients access this to request capability runs
    """
    @abstractmethod
    def run_capability(self, request: CapabilityRequestIF):
        """
        Run a capability with the requested settings
        :param request: Capability request
        """
        raise NotImplementedError


class EstimationServiceIF(ABC):
    """
    Service that estimates how long a capability will take to complete
    """
    pass


class MetricsServiceIF(ABC):
    """
    Service that records metrics
    """
    pass


class ArchiveServiceIF(ABC):
    """
    Abstracts services that are needed from the archive system.
    """
    @abstractmethod
    def lookup_product(self, locator: ProductLocator) -> "ProductIF":
        """
        Look up a science product by its locator
        :param locator:  science product locator for this product
        :return:         science product
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


FieldName = str
FieldLabel = str


"""

Capability System / Helpers

"""


class CapabilitySequenceIF(ABC):
    pass


class EventTranslatorIF(ABC):
    """
    Translates a Workflow event into a Capability event
    """
    pass


class CapabilityStepIF(ABC):
    """
    A step in a capability sequence
    """
    pass


class FutureCapabilityResultIF(ProductIF):
    """
    Class that represents a product that will in the future result from a completed
    capability execution
    """
    pass


class ParameterIF(ABC):
    """
    Class that represents a set of capability parameters, a required input for the correct
    and complete execution of a capability
    """
    def json(self) -> str:
        """
        Convert to json-formatted string
        :return: JSON string
        """
        raise NotImplementedError
